// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_STREAM_H
#define OB_STREAM_H

namespace OB
{

//
// Declare the Hastable template
//
template <class K, class T> class Hashtable;

//
// Forward declaration
//
class CodeConverterBase;

struct CodeConverters
{
    CodeConverterBase* charConverter;
    CodeConverterBase* wcharConverter;
};

//
// The InputStream class
//
class InputStream
{
    CodeConverters codeConverters_;
    const CORBA::Octet* start_;
    const CORBA::Octet* pos_;
    CORBA::ULong size_;
    bool swap_;

public: // required by some compilers

    struct ValueState
    {
        ValueState()
            : chunked(false), nestingLevel(0), chunkStart(0), chunkSize(0)
        {
        }

        ValueState(const ValueState& v)
            : chunked(v.chunked), nestingLevel(v.nestingLevel),
              chunkStart(v.chunkStart), chunkSize(v.chunkSize)
        {
        }

        bool chunked;
        CORBA::Long nestingLevel;
        const CORBA::Octet* chunkStart;
        CORBA::Long chunkSize;
    };

    struct ValueInfo
    {
        CORBA::Long tag;
        const CORBA::Octet* headerPos;
        const CORBA::Octet* dataPos;
        OB::StringSeq ids;
        ValueState state;
    };

private:

    ValueState valueState_;

    typedef Hashtable<const CORBA::Octet*, CORBA::TypeCode_ptr> ReadTCHistory;
    ReadTCHistory* readTCHistory_;

    typedef Hashtable<const CORBA::Octet*, CORBA::ValueBase*> ValueHistory;
    ValueHistory* valueHistory_;

    typedef Hashtable<const CORBA::Octet*, ValueInfo> ValueInfoHistory;
    ValueInfoHistory* valueInfoHistory_;

    InputStream(const InputStream&);
    InputStream& operator=(const InputStream&);

    CORBA::Long readLongImpl();
    void readChunk(ValueState&);
    void checkChunk();
    void readValueInfo(ValueInfo&);
    CORBA::ValueBase* createValue(const char*, const ValueInfo&);

    bool chunked() { return valueState_.chunked; }

public:

    InputStream(CodeConverters*, const Buffer&, CORBA::ULong, bool);
    InputStream(const Buffer&, CORBA::ULong, bool);
    InputStream(CodeConverters*, const CORBA::Octet*,
		CORBA::ULong, CORBA::ULong, bool);
    InputStream(const CORBA::Octet*, CORBA::ULong, CORBA::ULong, bool);
    ~InputStream();

    void _OB_readEndian();
    void _OB_pos(const CORBA::Octet* pos) { pos_ = pos; }
    const CORBA::Octet* _OB_pos() const { return pos_; }
    bool _OB_swap() const { return swap_; }
    void _OB_swap(bool s) { swap_ = s; }

    void _OB_skip(CORBA::ULong length)
    {
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + length > size_)
            throw CORBA::MARSHAL(MinorReadOverflow, CORBA::COMPLETED_NO);
#endif
        pos_ += length;
    }

    void _OB_skipAlign(CORBA::ULong align)
    {
        CORBA::ULong offs = (align - 1) - (((unsigned long)pos_ + (align - 1))
                            % align);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs > size_)
            throw CORBA::MARSHAL(MinorReadOverflow, CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
    }

    CORBA::Boolean read_boolean()
    {
        if(chunked())
            checkChunk();

#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + 1 > size_)
            throw CORBA::MARSHAL(MinorReadBooleanOverflow,
                                 CORBA::COMPLETED_NO);
#endif
        CORBA::Boolean val = *pos_++;
        return val;
    }

    CORBA::Octet read_octet()
    {
        if(chunked())
            checkChunk();

#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + 1 > size_)
            throw CORBA::MARSHAL(MinorReadOctetOverflow, CORBA::COMPLETED_NO);
#endif
        CORBA::Octet val = *pos_++;
        return val;
    }

    CORBA::Short read_short();
    CORBA::UShort read_ushort();
    CORBA::Long read_long();
    CORBA::ULong read_ulong();
    CORBA::LongLong read_longlong();
    CORBA::ULongLong read_ulonglong();
    CORBA::Float read_float();
    CORBA::Double read_double();
    CORBA::LongDouble read_longdouble();
    CORBA::Char read_char();
    CORBA::WChar read_wchar();
    char* read_string();
    wchar_t* read_wstring();
    CORBA::Object_ptr read_Object();
    CORBA::TypeCode_ptr read_TypeCode();
    CORBA::Principal_ptr read_Principal();
    CORBA::Any* read_any();
    CORBA::Fixed read_fixed();

    //
    // The string argument should be supplied except in special cases
    // (e.g., custom marshalling).
    //
    CORBA::ValueBase* read_Value(const char* = 0);

    //
    // This version of read_any is needed to avoid excessive copying
    // of inout parameters.
    //
    void read_any(CORBA::Any&);

    void read_boolean_array(CORBA::Boolean* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);

            if(chunked())
                checkChunk();

#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + length > size_)
                throw CORBA::MARSHAL(MinorReadBooleanArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            memcpy(value, pos_, length);
            pos_ += length;
        }
    }

    void read_char_array(CORBA::Char* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);

            if(chunked())
                checkChunk();

#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + length > size_)
                throw CORBA::MARSHAL(MinorReadCharArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            memcpy(value, pos_, length);
            pos_ += length;
        }
    }

    void read_octet_array(CORBA::Octet* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);

            if(chunked())
                checkChunk();

#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + length > size_)
                throw CORBA::MARSHAL(MinorReadOctetArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            memcpy(value, pos_, length);
            pos_ += length;
        }
    }

    void read_wchar_array(CORBA::WChar*, CORBA::ULong);
    void read_short_array(CORBA::Short*, CORBA::ULong);
    void read_ushort_array(CORBA::UShort*, CORBA::ULong);
    void read_long_array(CORBA::Long*, CORBA::ULong);
    void read_ulong_array(CORBA::ULong*, CORBA::ULong);
    void read_longlong_array(CORBA::LongLong*, CORBA::ULong);
    void read_ulonglong_array(CORBA::ULongLong*, CORBA::ULong);
    void read_float_array(CORBA::Float*, CORBA::ULong);
    void read_double_array(CORBA::Double*, CORBA::ULong);
    void read_longdouble_array(CORBA::LongDouble*, CORBA::ULong);

    void read_any_NoTypeCode(CORBA::Any&, CORBA::TypeCode_ptr);
};

//
// The OutputStream class
//
class OutputStream
{
    CodeConverters codeConverters_;
    CORBA::Octet* start_;
    CORBA::Octet* pos_;
    CORBA::ULong size_;
    CORBA::ULong origSize_;
    Buffer* buf_;

    //
    // For chunking
    //
    bool chunked_;
    CORBA::Long nestingLevel_;
    CORBA::ULong chunkPos_;
    bool endTag_;

    typedef Hashtable<CORBA::TypeCode_ptr, const CORBA::Octet*> WriteTCHistory;
    WriteTCHistory* writeTCHistory_;

    //
    // These types are needed for indirection in valuetype marshalling.
    // ValueHistory remembers the position at which a valuetype was
    // marshalled.  IdHistory remembers the position of a repository ID.
    // IdListHistory remembers the position of a list of repository IDs.
    //
    typedef Hashtable<CORBA::ValueBase*, CORBA::Long> ValueHistory;
    typedef Hashtable<const char*, CORBA::Long> IdHistory;
    typedef Hashtable<const char**, CORBA::Long> IdListHistory;
    ValueHistory* valueHistory_;
    IdHistory* idHistory_;
    IdListHistory* idListHistory_;

    OutputStream(const OutputStream&);
    OutputStream& operator=(const OutputStream&);

    bool chunked() { return chunked_; }

public:

    OutputStream(CodeConverters*, CORBA::ULong, CORBA::ULong = 0);
    OutputStream(CORBA::ULong, CORBA::ULong = 0);
    OutputStream(CodeConverters*, Buffer&, CORBA::ULong);
    OutputStream(Buffer&, CORBA::ULong);
    OutputStream(CodeConverters*, CORBA::Octet*, CORBA::ULong, CORBA::ULong);
    OutputStream(CORBA::Octet*, CORBA::ULong, CORBA::ULong);
    ~OutputStream();

    Buffer* _OB_buffer(CORBA::Boolean = CORBA::FALSE);
    CORBA::ULong _OB_count() const { return pos_ - start_; }
    void _OB_addCapacity(CORBA::ULong, CORBA::ULong = 0);


    void write_boolean(CORBA::Boolean value)
    {
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + 1 > size_)
            throw CORBA::MARSHAL(MinorWriteBooleanOverflow,
                                 CORBA::COMPLETED_NO);
#endif
        *pos_++ = value;
    }

    void write_octet(CORBA::Octet value)
    {
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + 1 > size_)
            throw CORBA::MARSHAL(MinorWriteOctetOverflow, CORBA::COMPLETED_NO);
#endif
        *pos_++ = value;
    }

    void write_short(CORBA::Short value)
    {
        CORBA::ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 2 > size_)
            throw CORBA::MARSHAL(MinorWriteShortOverflow, CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::Short*)pos_ = value;
        pos_ += 2;
    }

    void write_ushort(CORBA::UShort value)
    {
        CORBA::ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 2 > size_)
            throw CORBA::MARSHAL(MinorWriteUShortOverflow,
                                 CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::UShort*)pos_ = value;
        pos_ += 2;
    }

    void write_long(CORBA::Long value)
    {
        CORBA::ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 4 > size_)
            throw CORBA::MARSHAL(MinorWriteLongOverflow, CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::Long*)pos_ = value;
        pos_ += 4;
    }

    void write_ulong(CORBA::ULong value)
    {
        CORBA::ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 4 > size_)
            throw CORBA::MARSHAL(MinorWriteULongOverflow, CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::ULong*)pos_ = value;
        pos_ += 4;
    }

    void write_longlong(CORBA::LongLong value)
    {
        CORBA::ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 8 > size_)
            throw CORBA::MARSHAL(MinorWriteLongLongOverflow,
                                 CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::LongLong*)pos_ = value;
        pos_ += 8;
    }

    void write_ulonglong(CORBA::ULongLong value)
    {
        CORBA::ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 8 > size_)
            throw CORBA::MARSHAL(MinorWriteULongLongOverflow,
                                 CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::ULongLong*)pos_ = value;
        pos_ += 8;
    }

    void write_float(CORBA::Float value)
    {
        CORBA::ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 4 > size_)
            throw CORBA::MARSHAL(MinorWriteFloatOverflow, CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::Float*)pos_ = value;
        pos_ += 4;
    }

    void write_double(CORBA::Double value)
    {
        CORBA::ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
        if((((unsigned long)(pos_ - start_))) + offs + 8 > size_)
            throw CORBA::MARSHAL(MinorWriteDoubleOverflow,
                                 CORBA::COMPLETED_NO);
#endif
        pos_ += offs;
        *(CORBA::Double*)pos_ = value;
        pos_ += 8;
    }

    void write_longdouble(CORBA::LongDouble);
    void write_char(CORBA::Char);
    void write_wchar(CORBA::WChar);
    void write_string(const char*);
    void write_wstring(const wchar_t*);
    void write_Object(CORBA::Object_ptr);
    void write_TypeCode(CORBA::TypeCode_ptr);
    void write_Principal(CORBA::Principal_ptr);
    void write_any(const CORBA::Any&);
    void write_fixed(const CORBA::Fixed&);

    //
    // The string argument should be supplied except in special cases
    // (e.g., custom marshalling).
    //
    void write_Value(CORBA::ValueBase*, const char* = 0);

    void write_Abstract(CORBA::AbstractBase_ptr);

    void write_boolean_array(const CORBA::Boolean* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + length > size_)
                throw CORBA::MARSHAL(MinorWriteBooleanArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            memcpy(pos_, value, length);
            pos_ += length;
        }
    }

    void write_char_array(const CORBA::Char* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + length > size_)
                throw CORBA::MARSHAL(MinorWriteCharArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            memcpy(pos_, value, length);
            pos_ += length;
        }
    }

    void write_octet_array(const CORBA::Octet* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + length > size_)
                throw CORBA::MARSHAL(MinorWriteOctetArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            memcpy(pos_, value, length);
            pos_ += length;
        }
    }

    void write_short_array(const CORBA::Short* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 2 > size_)
                throw CORBA::MARSHAL(MinorWriteShortArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 2);
            pos_ += length * 2;
        }
    }

    void write_ushort_array(const CORBA::UShort* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = (unsigned long)pos_ % 2;
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 2 > size_)
                throw CORBA::MARSHAL(MinorWriteUShortArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 2);
            pos_ += length * 2;
        }
    }

    void write_long_array(const CORBA::Long* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 4 > size_)
                throw CORBA::MARSHAL(MinorWriteLongArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 4);
            pos_ += length * 4;
        }
    }

    void write_ulong_array(const CORBA::ULong* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 4 > size_)
                throw CORBA::MARSHAL(MinorWriteULongArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 4);
            pos_ += length * 4;
        }
    }

    void write_longlong_array(const CORBA::LongLong* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 8 > size_)
                throw CORBA::MARSHAL(MinorWriteLongLongArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 8);
            pos_ += length * 8;
        }
    }

    void write_ulonglong_array(const CORBA::ULongLong* value,
                               CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 8 > size_)
                throw CORBA::MARSHAL(MinorWriteULongLongArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 8);
            pos_ += length * 8;
        }
    }

    void write_float_array(const CORBA::Float* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = 3 - (((unsigned long)pos_ + 3) % 4);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 4 > size_)
                throw CORBA::MARSHAL(MinorWriteFloatArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 4);
            pos_ += length * 4;
        }
    }

    void write_double_array(const CORBA::Double* value, CORBA::ULong length)
    {
        if(length)
        {
            assert_nca(value, NCANullValue);
            CORBA::ULong offs = 7 - (((unsigned long)pos_ + 7) % 8);
#ifdef OB_MARSHAL_CHECK
            if((((unsigned long)(pos_ - start_))) + offs + length * 8 > size_)
                throw CORBA::MARSHAL(MinorWriteDoubleArrayOverflow,
                                     CORBA::COMPLETED_NO);
#endif
            pos_ += offs;
            memcpy(pos_, value, length * 8);
            pos_ += length * 8;
        }
    }

    void write_wchar_array(const CORBA::WChar*, CORBA::ULong);
    void write_longdouble_array(const CORBA::LongDouble*, CORBA::ULong);

    void write_any_NoTypeCode(const CORBA::Any&);
    void write_InputStream(CORBA::TypeCode_ptr, InputStream&);
};

//
// The CountStream class
//
class CountStream
{
    CodeConverters codeConverters_;
    CORBA::ULong count_;

    //
    // For chunking
    //
    bool chunked_;
    CORBA::Long nestingLevel_;
    CORBA::ULong chunkPos_;
    bool endTag_;

    typedef Hashtable<CORBA::TypeCode_ptr, const CORBA::Octet*> CountTCHistory;
    CountTCHistory* countTCHistory_;

    //
    // These types are needed for indirection in valuetype marshalling.
    // ValueHistory remembers the position at which a valuetype was
    // marshalled.  IdHistory remembers the position of a repository ID.
    // IdListHistory remembers the position of a list of repository IDs.
    //
    typedef Hashtable<CORBA::ValueBase*, CORBA::Long> ValueHistory;
    typedef Hashtable<const char*, CORBA::Long> IdHistory;
    typedef Hashtable<const char**, CORBA::Long> IdListHistory;
    ValueHistory* valueHistory_;
    IdHistory* idHistory_;
    IdListHistory* idListHistory_;

    CountStream(const CountStream&);
    CountStream& operator=(const CountStream&);

    friend class OutputStream;

    bool chunked() { return chunked_; }

public:

    CountStream(CodeConverters*);
    CountStream();
    CountStream(CodeConverters*, CORBA::ULong);
    CountStream(CORBA::ULong);
    ~CountStream();

    CORBA::ULong count() const { return count_; }

    void add(CORBA::ULong size, CORBA::ULong align = 0)
    {
        if(align > 0)
        {
            count_ += align - 1;
            count_ -= count_ % align;
        }
        count_ += size;
    }

    void count_boolean()
    {
        count_++;
    }

    void count_char()
    {
        count_++;
    }

    void count_octet()
    {
        count_++;
    }

    void count_short()
    {
        count_ += count_ % 2 + 2;
    }

    void count_ushort()
    {
        count_ += count_ % 2 + 2;
        count_ += 2;
    }

    void count_long()
    {
        count_ += 3;
        count_ -= count_ % 4;
        count_ += 4;
    }

    void count_ulong()
    {
        count_ += 3;
        count_ -= count_ % 4;
        count_ += 4;
    }

    void count_longlong()
    {
        count_ += 7;
        count_ -= count_ % 8;
        count_ += 8;
    }

    void count_ulonglong()
    {
        count_ += 7;
        count_ -= count_ % 8;
        count_ += 8;
    }

    void count_float()
    {
        count_ += 3;
        count_ -= count_ % 4;
        count_ += 4;
    }

    void count_double()
    {
        count_ += 7;
        count_ -= count_ % 8;
        count_ += 8;
    }

    void count_longdouble()
    {
        count_ += 7;
        count_ -= count_ % 8;
        count_ += 16;
    }

    void count_string(const char* value)
    {
	assert_nca(value, NCANullString);
	CORBA::ULong len = strlen(value) + 1;
	count_ulong(); // len
	count_ += len;
    }

    void count_wchar(CORBA::WChar);
    void count_wstring(const wchar_t*);
    void count_Object(CORBA::Object_ptr);
    void count_TypeCode(CORBA::TypeCode_ptr);
    void count_Principal(CORBA::Principal_ptr);
    void count_any(const CORBA::Any&);
    void count_fixed(const CORBA::Fixed&);

    //
    // The string argument should be supplied except in special cases
    // (e.g., custom marshalling).
    //
    void count_Value(CORBA::ValueBase*, const char* = 0);

    void count_Abstract(CORBA::AbstractBase_ptr);

    void count_boolean_array(CORBA::ULong length)
    {
        count_ += length;
    }

    void count_char_array(CORBA::ULong length)
    {
        count_ += length;
    }

    void count_octet_array(CORBA::ULong length)
    {
        count_ += length;
    }

    void count_short_array(CORBA::ULong length)
    {
        if(length)
            count_ += count_ % 2 + length * 2;
    }

    void count_ushort_array(CORBA::ULong length)
    {
        if(length)
            count_ += count_ % 2 + length * 2;
    }

    void count_long_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 3;
            count_ -= count_ % 4;
            count_ += length * 4;
        }
    }

    void count_ulong_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 3;
            count_ -= count_ % 4;
            count_ += length * 4;
        }
    }

    void count_longlong_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 7;
            count_ -= count_ % 8;
            count_ += length * 8;
        }
    }

    void count_ulonglong_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 7;
            count_ -= count_ % 8;
            count_ += length * 8;
        }
    }

    void count_float_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 3;
            count_ -= count_ % 4;
            count_ += length * 4;
        }
    }

    void count_double_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 7;
            count_ -= count_ % 8;
            count_ += length * 8;
        }
    }

    void count_longdouble_array(CORBA::ULong length)
    {
        if(length)
        {
            count_ += 7;
            count_ -= count_ % 8;
            count_ += length * 16;
        }
    }

    void count_wchar_array(const CORBA::WChar*, CORBA::ULong);

    void count_any_NoTypeCode(const CORBA::Any&);

    void count_InputStream(CORBA::TypeCode_ptr, InputStream&);
};

} // End of namespace OB

#endif
